@@ -300,7 +300,9 @@ addToLibrary({
             var encoder = new TextEncoder(); // should be utf-8
             data = encoder.encode(data); // make a typed array from the string
           } else {
+#if ASSERTIONS
             assert(data.byteLength !== undefined); // must receive an ArrayBuffer
+#endif
             if (data.byteLength == 0) {
               // An empty ArrayBuffer will emit a pseudo disconnect event
               // as recv/recvmsg will return zero which indicates that a socket
@@ -421,6 +423,14 @@ addToLibrary({
             }
             {{{ makeSetValue('arg', '0', 'bytes', 'i32') }}};
             return 0;
+          case {{{ cDefs.FIONBIO }}}:
+            var on = {{{ makeGetValue('arg', '0', 'i32') }}};
+            if (on) {
+              sock.stream.flags |= {{{ cDefs.O_NONBLOCK }}};
+            } else {
+              sock.stream.flags &= ~{{{ cDefs.O_NONBLOCK }}};
+            }
+            return 0;
           default:
             return {{{ cDefs.EINVAL }}};
         }
