@@ -36,9 +36,11 @@ namespace internal {
 template<typename WrapperType>
 val wrapped_extend(const std::string&, const val&);
 
-enum class EM_METHOD_CALLER_KIND {
-  FUNCTION = 0,
-  CONSTRUCTOR = 1,
+enum class EM_INVOKER_KIND {
+  FUNCTION,
+  METHOD,
+  CONSTRUCTOR,
+  CAST,
 };
 
 // Implemented in JavaScript.  Don't call these directly.
@@ -55,7 +57,7 @@ enum {
 };
 
 typedef struct _EM_DESTRUCTORS* EM_DESTRUCTORS;
-typedef struct _EM_METHOD_CALLER* EM_METHOD_CALLER;
+typedef struct _EM_INVOKER* EM_INVOKER;
 typedef double EM_GENERIC_WIRE_TYPE;
 typedef const void* EM_VAR_ARGS;
 
@@ -71,15 +73,10 @@ EM_VAL _emval_new_cstring(const char*);
 EM_VAL _emval_new_u8string(const char*);
 EM_VAL _emval_new_u16string(const char16_t*);
 
-EM_VAL _emval_take_value(TYPEID type, EM_VAR_ARGS argv);
-
 EM_VAL _emval_get_global(const char* name);
 EM_VAL _emval_get_module_property(const char* name);
 EM_VAL _emval_get_property(EM_VAL object, EM_VAL key);
 void _emval_set_property(EM_VAL object, EM_VAL key, EM_VAL value);
-EM_GENERIC_WIRE_TYPE _emval_as(EM_VAL value, TYPEID returnType, EM_DESTRUCTORS* destructors);
-int64_t _emval_as_int64(EM_VAL value, TYPEID returnType);
-uint64_t _emval_as_uint64(EM_VAL value, TYPEID returnType);
 
 bool _emval_equals(EM_VAL first, EM_VAL second);
 bool _emval_strictly_equals(EM_VAL first, EM_VAL second);
@@ -87,20 +84,20 @@ bool _emval_greater_than(EM_VAL first, EM_VAL second);
 bool _emval_less_than(EM_VAL first, EM_VAL second);
 bool _emval_not(EM_VAL object);
 
-EM_GENERIC_WIRE_TYPE _emval_call(
-    EM_METHOD_CALLER caller,
-    EM_VAL func,
-    EM_DESTRUCTORS* destructors,
-    EM_VAR_ARGS argv);
-
 // DO NOT call this more than once per signature. It will
 // leak generated function objects!
-EM_METHOD_CALLER _emval_get_method_caller(
+EM_INVOKER _emval_create_invoker(
     unsigned argCount, // including return value
     const TYPEID argTypes[],
-    EM_METHOD_CALLER_KIND asCtor);
-EM_GENERIC_WIRE_TYPE _emval_call_method(
-    EM_METHOD_CALLER caller,
+    EM_INVOKER_KIND kind);
+EM_GENERIC_WIRE_TYPE _emval_invoke(
+    EM_INVOKER caller,
+    EM_VAL handle,
+    const char* methodName,
+    EM_DESTRUCTORS* destructors,
+    EM_VAR_ARGS argv);
+int64_t _emval_invoke_i64(
+    EM_INVOKER caller,
     EM_VAL handle,
     const char* methodName,
     EM_DESTRUCTORS* destructors,
@@ -131,22 +128,6 @@ struct symbol_registrar {
   }
 };
 
-template<EM_METHOD_CALLER_KIND Kind, typename ReturnType, typename... Args>
-struct Signature {
-  /*
-  typedef typename BindingType<ReturnType>::WireType (*MethodCaller)(
-      EM_VAL object,
-      EM_VAL method,
-      EM_DESTRUCTORS* destructors,
-      typename BindingType<Args>::WireType...);
-  */
-  static EM_METHOD_CALLER get_method_caller() {
-    static constexpr WithPolicies<>::ArgTypeList<ReturnType, Args...> args;
-    thread_local EM_METHOD_CALLER mc = _emval_get_method_caller(args.getCount(), args.getTypes(), Kind);
-    return mc;
-  }
-};
-
 struct DestructorsRunner {
 public:
   explicit DestructorsRunner(EM_DESTRUCTORS d)
@@ -179,17 +160,12 @@ struct GenericWireTypeConverter<Pointee*> {
   }
 };
 
-template<typename T>
-T fromGenericWireType(EM_GENERIC_WIRE_TYPE g) {
-  typedef typename BindingType<T>::WireType WireType;
-  WireType wt = GenericWireTypeConverter<WireType>::from(g);
-  return BindingType<T>::fromWireType(wt);
-}
-
 template<>
-inline void fromGenericWireType<void>(EM_GENERIC_WIRE_TYPE g) {
-  (void)g;
-}
+struct GenericWireTypeConverter<BindingType<void>::WireType> {
+  static BindingType<void>::WireType from(double) {
+    return {};
+  }
+};
 
 template<typename... Args>
 struct PackSize;
@@ -368,9 +344,8 @@ public:
   template<typename T, typename... Policies>
   explicit val(T&& value, Policies...) {
     using namespace internal;
-    typename WithPolicies<Policies...>::template ArgTypeList<T> valueType;
-    WireTypePack<T> argv(std::forward<T>(value));
-    new (this) val(_emval_take_value(valueType.getTypes()[0], argv));
+
+    new (this) val(internalCall<EM_INVOKER_KIND::CAST, WithPolicies<Policies...>, val>(nullptr, nullptr, std::forward<T>(value)));
   }
 
   val() : val(EM_VAL(internal::_EMVAL_UNDEFINED)) {}
@@ -393,6 +368,11 @@ public:
     }
   }
 
+  // Add an explicit overload for `val&` as well.
+  // Without it, C++ will try to use the `T&&` constructor instead of the more
+  // efficient `val(const val&)` when trying to copy a `val` instance.
+  val(val& v) : val(static_cast<const val&>(v)) {}
+
   ~val() {
     if (uses_ref_count()) {
       internal::_emval_decref(as_handle());
@@ -498,9 +478,9 @@ public:
     return val(internal::_emval_get_property(as_handle(), val_ref(key).as_handle()));
   }
 
-  template<typename K, typename V>
-  void set(const K& key, const V& value) {
-    internal::_emval_set_property(as_handle(), val_ref(key).as_handle(), val_ref(value).as_handle());
+  template<typename K, typename V, typename... Policies>
+  void set(const K& key, const V& value, Policies... policies) {
+    internal::_emval_set_property(as_handle(), val_ref(key).as_handle(), val_ref(value, policies...).as_handle());
   }
 
   template<typename T>
@@ -512,86 +492,28 @@ public:
   val new_(Args&&... args) const {
     using namespace internal;
 
-    return internalCall<EM_METHOD_CALLER_KIND::CONSTRUCTOR, val>(_emval_call, std::forward<Args>(args)...);
+    return internalCall<EM_INVOKER_KIND::CONSTRUCTOR, WithPolicies<>, val>(as_handle(), nullptr, std::forward<Args>(args)...);
   }
 
   template<typename... Args>
   val operator()(Args&&... args) const {
     using namespace internal;
 
-    return internalCall<EM_METHOD_CALLER_KIND::FUNCTION, val>(_emval_call, std::forward<Args>(args)...);
+    return internalCall<EM_INVOKER_KIND::FUNCTION, WithPolicies<>, val>(as_handle(), nullptr, std::forward<Args>(args)...);
   }
 
   template<typename ReturnValue, typename... Args>
   ReturnValue call(const char* name, Args&&... args) const {
     using namespace internal;
 
-    return internalCall<EM_METHOD_CALLER_KIND::FUNCTION, ReturnValue>(
-      [name](EM_METHOD_CALLER caller,
-             EM_VAL handle,
-             EM_DESTRUCTORS* destructorsRef,
-             EM_VAR_ARGS argv) {
-        return _emval_call_method(caller, handle, name, destructorsRef, argv);
-      },
-      std::forward<Args>(args)...);
+    return internalCall<EM_INVOKER_KIND::METHOD, WithPolicies<>, ReturnValue>(as_handle(), name, std::forward<Args>(args)...);
   }
 
   template<typename T, typename ...Policies>
   T as(Policies...) const {
     using namespace internal;
 
-    typedef BindingType<T> BT;
-    typename WithPolicies<Policies...>::template ArgTypeList<T> targetType;
-
-    EM_DESTRUCTORS destructors = nullptr;
-    EM_GENERIC_WIRE_TYPE result = _emval_as(
-        as_handle(),
-        targetType.getTypes()[0],
-        &destructors);
-    DestructorsRunner dr(destructors);
-    return fromGenericWireType<T>(result);
-  }
-
-#ifdef __wasm64__
-  template<>
-  long as<long>() const {
-    using namespace internal;
-
-    typedef BindingType<long> BT;
-    typename WithPolicies<>::template ArgTypeList<long> targetType;
-
-    return _emval_as_int64(as_handle(), targetType.getTypes()[0]);
-  }
-
-  template<>
-  unsigned long as<unsigned long>() const {
-    using namespace internal;
-
-    typedef BindingType<unsigned long> BT;
-    typename WithPolicies<>::template ArgTypeList<unsigned long> targetType;
-
-    return _emval_as_uint64(as_handle(), targetType.getTypes()[0]);
-  }
-#endif
-
-  template<>
-  int64_t as<int64_t>() const {
-    using namespace internal;
-
-    typedef BindingType<int64_t> BT;
-    typename WithPolicies<>::template ArgTypeList<int64_t> targetType;
-
-    return _emval_as_int64(as_handle(), targetType.getTypes()[0]);
-  }
-
-  template<>
-  uint64_t as<uint64_t>() const {
-    using namespace internal;
-
-    typedef BindingType<uint64_t> BT;
-    typename WithPolicies<>::template ArgTypeList<uint64_t> targetType;
-
-    return  _emval_as_uint64(as_handle(), targetType.getTypes()[0]);
+    return internalCall<EM_INVOKER_KIND::CAST, WithPolicies<Policies...>, T>(as_handle(), nullptr, *this);
   }
 
 // Prefer calling val::typeOf() over val::typeof(), since this form works in both C++11 and GNU++11 build modes. "typeof" is a reserved word in GNU++11 extensions.
@@ -651,24 +573,45 @@ private:
   template<typename WrapperType>
   friend val internal::wrapped_extend(const std::string& , const val& );
 
-  template<internal::EM_METHOD_CALLER_KIND Kind, typename Ret, typename Implementation, typename... Args>
-  Ret internalCall(Implementation impl, Args&&... args) const {
+  template<internal::EM_INVOKER_KIND Kind, typename Policy, typename Ret, typename... Args>
+  static Ret internalCall(EM_VAL handle, const char *methodName, Args&&... args) {
+    static_assert(!std::is_lvalue_reference<Ret>::value,
+                  "Cannot create a lvalue reference out of a JS value.");
+
     using namespace internal;
 
+    using RetWire = BindingType<Ret>::WireType;
+
+    static constexpr typename Policy::template ArgTypeList<Ret, Args...> argTypes;
+    thread_local EM_INVOKER mc = _emval_create_invoker(argTypes.getCount(), argTypes.getTypes(), Kind);
+
     WireTypePack<Args...> argv(std::forward<Args>(args)...);
     EM_DESTRUCTORS destructors = nullptr;
-    EM_GENERIC_WIRE_TYPE result = impl(
-      Signature<Kind, Ret, Args...>::get_method_caller(),
-      as_handle(),
-      &destructors,
-      argv);
+
+    RetWire result;
+    if constexpr (std::is_integral<RetWire>::value && sizeof(RetWire) == 8) {
+      // 64-bit integers can't go through "generic wire type" because double and int64 have different ABI.
+      result = static_cast<RetWire>(_emval_invoke_i64(
+        mc,
+        handle,
+        methodName,
+        &destructors,
+        argv));
+    } else {
+      result = GenericWireTypeConverter<RetWire>::from(_emval_invoke(
+        mc,
+        handle,
+        methodName,
+        &destructors,
+        argv));
+    }
     DestructorsRunner rd(destructors);
-    return fromGenericWireType<Ret>(result);
+    return BindingType<Ret>::fromWireType(result);
   }
 
-  template<typename T>
-  val val_ref(const T& v) const {
-    return val(v);
+  template<typename T, typename... Policies>
+  val val_ref(const T& v, Policies... policies) const {
+    return val(v, policies...);
   }
 
   const val& val_ref(const val& v) const {
@@ -785,6 +728,9 @@ public:
   auto initial_suspend() noexcept { return std::suspend_never{}; }
   auto final_suspend() noexcept { return std::suspend_never{}; }
 
+// When exceptions are disabled we don't define unhandled_exception and rely
+// on the default terminate behavior.
+#ifdef __cpp_exceptions
   // On an unhandled exception, reject the stored promise instead of throwing
   // it asynchronously where it can't be handled.
   void unhandled_exception() {
@@ -797,6 +743,7 @@ public:
       reject(error);
     }
   }
+#endif
 
   // Reject the stored promise due to rejection deeper in the call chain
   void reject_with(val&& error) {
