@@ -1,5 +1,5 @@
 addEventListener('error', (event) => {
   window.disableErrorReporting = true;
   var result = event.error === 42 ? 0 : 1;
-  fetch('http://localhost:8888/report_result?' + result);
+  fetch('/report_result?' + result);
 });
