@@ -1081,24 +1081,40 @@ def instrument_js_for_safe_heap(js_file):
   return acorn_optimizer(js_file, ['safeHeap'])
 
 
+def read_name_section(wasm_file):
+  with webassembly.Module(wasm_file) as module:
+    for section in module.sections():
+      if section.type == webassembly.SecType.CUSTOM:
+        module.seek(section.offset)
+        if module.read_string() == 'name':
+          name_map = {}
+          # The name section is made up sub-section.
+          # We are looking for the function names sub-section
+          while module.tell() < section.offset + section.size:
+            name_type = module.read_uleb()
+            subsection_size = module.read_uleb()
+            subsection_end = module.tell() + subsection_size
+            if name_type == webassembly.NameType.FUNCTION:
+              # We found the function names sub-section
+              num_names = module.read_uleb()
+              for _ in range(num_names):
+                id = module.read_uleb()
+                name = module.read_string()
+                name_map[id] = name
+              return name_map
+            module.seek(subsection_end)
+
+          return name_map
+
+
 @ToolchainProfiler.profile()
-def handle_final_wasm_symbols(wasm_file, symbols_file, debug_info):
+def write_symbol_map(wasm_file, symbols_file):
   logger.debug('handle_final_wasm_symbols')
-  args = []
-  if symbols_file:
-    args += ['--print-function-map']
-  else:
-    # suppress the wasm-opt warning regarding "no output file specified"
-    args += ['--quiet']
-  output = run_wasm_opt(wasm_file, args=args, stdout=PIPE)
-  if symbols_file:
-    utils.write_file(symbols_file, output)
-  if not debug_info:
-    # strip the names section using llvm-objcopy. this is slightly slower than
-    # using wasm-opt (we could run wasm-opt without -g here and just tell it to
-    # write the file back out), but running wasm-opt would undo StackIR
-    # optimizations, if we did those.
-    strip(wasm_file, wasm_file, sections=['name'])
+  names = read_name_section(wasm_file)
+  assert(names)
+  strings = [f'{id}:{name}' for id, name in names.items()]
+  contents = '\n'.join(strings) + '\n'
+  utils.write_file(symbols_file, contents)
 
 
 def is_ar(filename):
