@@ -16,7 +16,6 @@ import json
 import subprocess
 import logging
 import pprint
-import re
 import shutil
 import sys
 import textwrap
@@ -318,26 +317,15 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
     update_settings_glue(out_wasm, metadata, base_metadata)
 
   if not settings.WASM_BIGINT and metadata.em_js_funcs:
-    import_map = {}
-
-    with webassembly.Module(in_wasm) as module:
-      types = module.get_types()
-      for imp in module.get_imports():
-        if imp.module not in ('GOT.mem', 'GOT.func'):
-          import_map[imp.field] = imp
-
     for em_js_func, raw in metadata.em_js_funcs.items():
       c_sig = raw.split('<::>')[0].strip('()')
       if not c_sig or c_sig == 'void':
         c_sig = []
       else:
         c_sig = c_sig.split(',')
-      if em_js_func in import_map:
-        imp = import_map[em_js_func]
-        assert imp.kind == webassembly.ExternType.FUNC
-        signature = types[imp.type]
-        if len(signature.params) != len(c_sig):
-          diagnostics.warning('em-js-i64', 'using 64-bit arguments in EM_JS function without WASM_BIGINT is not yet fully supported: `%s` (%s, %s)', em_js_func, c_sig, signature.params)
+      signature = metadata.em_js_func_types.get(em_js_func)
+      if signature and len(signature.params) != len(c_sig):
+        diagnostics.warning('em-js-i64', 'using 64-bit arguments in EM_JS function without WASM_BIGINT is not yet fully supported: `%s` (%s, %s)', em_js_func, c_sig, signature.params)
 
   asm_consts = create_asm_consts(metadata)
   em_js_funcs = create_em_js(metadata)
@@ -365,6 +353,10 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
     logger.debug('emscript: skipping js glue generation')
     return
 
+  for e in settings.EXPORTED_FUNCTIONS:
+    if not js_manipulation.isidentifier(e):
+      exit_with_error(f'invalid export name: "{e}"')
+
   # memory and global initializers
 
   if settings.RELOCATABLE:
@@ -456,13 +448,6 @@ def get_metadata(infile, outfile, modify_wasm, args):
   return metadata
 
 
-def is_valid_js_identifier(ident):
-  # See https://developer.mozilla.org/en-US/docs/Glossary/Identifier
-  if ident[0].isdigit():
-    return False
-  return re.fullmatch(r'[0-9a-zA-Z_\$]+', ident)
-
-
 def finalize_wasm(infile, outfile, js_syms):
   building.save_intermediate(infile, 'base.wasm')
   args = []
@@ -570,9 +555,6 @@ def finalize_wasm(infile, outfile, js_syms):
   # These are any exports that were not requested on the command line and are
   # not known auto-generated system functions.
   unexpected_exports = [e for e in metadata.all_exports if shared.is_user_export(e)]
-  for n in unexpected_exports:
-    if not is_valid_js_identifier(n):
-      exit_with_error(f'invalid export name: {n}')
   unexpected_exports = [asmjs_mangle(e) for e in unexpected_exports]
   unexpected_exports = [e for e in unexpected_exports if e not in expected_exports]
 
@@ -915,18 +897,15 @@ def install_debug_wrapper(sym):
     return False
   # Likewise `__trap` can occur before the runtime is initialized since it is used in
   # abort.
-  # pthread_self and _emscripten_proxy_execute_task_queue are currently called in some
-  # cases after the runtime has exited.
-  # TODO: Look into removing these, and improving our robustness around thread termination.
-  return sym not in {'__trap', 'pthread_self', '_emscripten_proxy_execute_task_queue'}
+  return sym != '__trap'
 
 
 def should_export(sym):
   return settings.EXPORT_ALL or (settings.EXPORT_KEEPALIVE and sym in settings.EXPORTED_FUNCTIONS)
 
 
-def create_receiving(function_exports, tag_exports):
-  generate_dyncall_assignment = settings.DYNCALLS and '$dynCall' in settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE
+def create_receiving(function_exports, tag_exports, library_symbols):
+  generate_dyncall_assignment = 'dynCalls' in library_symbols
   receiving = ['\n// Imports from the Wasm binary.']
 
   if settings.WASM_ESM_INTEGRATION:
@@ -1013,10 +992,10 @@ def create_receiving(function_exports, tag_exports):
   return '\n'.join(receiving) + '\n'
 
 
-def create_module(metadata, function_exports, global_exports, tag_exports,library_symbols):
+def create_module(metadata, function_exports, global_exports, tag_exports, library_symbols):
   module = []
 
-  receiving = create_receiving(function_exports, tag_exports)
+  receiving = create_receiving(function_exports, tag_exports, library_symbols)
   receiving += create_global_exports(global_exports)
   sending = create_sending(metadata, library_symbols)
 
@@ -1038,21 +1017,6 @@ def create_module(metadata, function_exports, global_exports, tag_exports,librar
     else:
       module.append('var wasmImports = %s;\n' % sending)
 
-    if not settings.MINIMAL_RUNTIME:
-      if settings.MODULARIZE == 'instance':
-        module.append("var wasmExports;\n")
-      elif settings.WASM_ASYNC_COMPILATION:
-        if settings.MODULARIZE:
-          # In modularize mode the generated code is within a factory function.
-          # This magic string gets replaced by `await createWasm`.  It needed to allow
-          # closure and acorn to process the module without seeing this as a top-level
-          # await.
-          module.append("var wasmExports = EMSCRIPTEN$AWAIT(createWasm());\n")
-        else:
-          module.append("var wasmExports;\ncreateWasm();\n")
-      else:
-        module.append("var wasmExports = createWasm();\n")
-
   if settings.SUPPORT_LONGJMP == 'emscripten' or not settings.DISABLE_EXCEPTION_CATCHING:
     module.append(create_invoke_wrappers(metadata))
   else:
@@ -1131,8 +1095,7 @@ def create_pointer_conversion_wrappers(metadata):
     '_wasmfs_read_file': '_ppp',
     '_wasmfs_node_record_dirent': '_pp_',
     '__dl_seterr': '_pp',
-    '_emscripten_run_on_main_thread_js': '__p_p_',
-    '_emscripten_proxy_execute_task_queue': '_p',
+    '_emscripten_run_js_on_main_thread': '__p_p_',
     '_emscripten_thread_exit': '_p',
     '_emscripten_thread_init': '_p_____',
     '_emscripten_thread_free_data': '_p',
